#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include "wavereader.h"

static void giveup (char *why);
static void usage ();
static float* carica_impulso (WaveFileDirect *wfh);

static void fai_convoluzine (WaveFileDirect *wfx, /* file di input */
							 float *impulso, int campioni_impulso, int canali_impulso, /* impulso */
							 WaveFileWriter *wfo); /* file di output */

/*-------------------------------------------------------------------------------

	Sintassi del programma:

		conv audio.wav impulso.wav output.wav

-------------------------------------------------------------------------------*/
int main (int argc, char** argv)
{
	char* file_input;
	char* file_impulso;
	char* file_output;

	WaveFileDirect* wfx; /* file dati */	
	WaveFileDirect* wfh; /* file impulso */
	WaveFileWriter* wfo; /* file scrittura */

	int canali_output; /* numero di canali per output */
	int canali_input;  /* numero di canali per input */
	int canali_impulso; /* numero di canali per l'impulso */

	float* impulso;
	int campioni_impulso; /* numero di campioni del file d'impulso */
	

	if (argc != 4)
		usage ();

	/* nomi dei file passati alla line di comando */
	file_input = argv[1];
	file_impulso = argv[2];
	file_output = argv[3];

	/* apre file di input */
	wfx = WFD_ApriWave (file_input);
	if (wfx == NULL) 
		giveup ("Impossibile aprire il file di input");

	wfh = WFD_ApriWave(file_impulso);
	if (wfh == NULL)
		giveup ("Impossibile aprire il file dell'impulso");

	if (wfx->sampleRate != wfh->sampleRate)
		giveup("I file di input e di impulso devono avere la stessa frequenza di campionamento");

	canali_input = wfx->nChannels;
	canali_impulso = wfh->nChannels;
	campioni_impulso = wfh->nSamples;

	/* se almeno uno dei due file  stereo allora generiamo un file
	   stereo, altrimenti lasciamo il tutto in mono */
	if (canali_input == 2 || canali_impulso == 2)
		canali_output = 2;
	else
		canali_output = 1;

	/* crea il file wave di output... */
	wfo = WFW_CreaWave(file_output, wfx->sampleRate, canali_output);
	if (wfo == NULL)
		giveup("Impossibile creare il file di output");

	impulso = carica_impulso(wfh);
	if (impulso == NULL)
	{	
		giveup("Impossibile caricare l'impulso");
	}

	fai_convoluzine(wfx, impulso, campioni_impulso, canali_impulso, wfo);

	WFW_Chiudi(wfo);
	WFD_Chiudi(wfx);
	WFD_Chiudi(wfh);
	free (impulso);

	return 0;

}


void fai_convoluzine(WaveFileDirect *wfx, 
					 float *impulso, 
					 int campioni_impulso,
					 int canali_impulso,
					 WaveFileWriter *wfo)
{
	int j, k, l;
	float  a, b;

	/* Ribalta l'impulso simmetricamente; 
	   l'algoritmo e' quello gia' visto nella lezione n. ?? applicato
	   alle stringhe di caratteri */
	if (canali_impulso == 1)
		for (k = 0; k < campioni_impulso/2; ++k)
		{
			l = campioni_impulso - 1 - k;
			a = impulso[k];
			impulso[k] = impulso[l];
			impulso[l] = a;
		}
	else 
		for (k = 0; k < campioni_impulso/2; ++k)
		{			
			l = campioni_impulso - 1 - k;
			a = impulso[k*2];
			b = impulso[k*2+1];
			impulso[k*2] = impulso[l*2];
			impulso[k*2+1] = impulso[l*2+1];
			impulso[l*2] = a;
			impulso[l*2+1] = b;
		}

	/* 3- effettua la convoluzione (commenti solo in questa funzione) */
	
	/* x(n) mono, h(n) mono */
	if (wfx->nChannels == 1 && canali_impulso == 1)
	{
		int campioni_da_leggere = wfx->nSamples;
		float  campione_input, 
			   campione_output;

		/* crea buffer di delay per contenere i campioni del segnale x(n) necessari 
		   per la convoluzione */
		float* delay_buffer = malloc(sizeof(float)*campioni_impulso);
		memset (delay_buffer, 0, sizeof(float)*campioni_impulso);

		/* per ciascun campione di x(n)... */
		while (campioni_da_leggere--)
		{
			/* leggi campione dal file */
			WFD_LeggiCampione(wfx, &campione_input);

			/* somma il prodotto tra x(n) e h(n) */
			campione_output = delay_buffer[0] * impulso[0];
			for (k = 1; k < campioni_impulso; ++k)
			{	
				/* MAC */
				campione_output += delay_buffer[k] * impulso[k];

				/* sposta a sinistra di un campione il delay */
				delay_buffer[k-1] = delay_buffer[k]; 
			}

			/* aggiunge il nuovo campione in coda */
			delay_buffer[k-1] = campione_input; 

			/* scrive il campione risultante su file */
			WFW_Scrivi(wfo, &campione_output, 1);
		}
		
		/* completa la coda dell'impulso */
		for (j = 0; j < campioni_impulso; ++j)
		{
			campione_output = delay_buffer[0] * impulso[0];
			for (k = 1; k < campioni_impulso; ++k)
			{	
				/* MAC */
				campione_output += delay_buffer[k] * impulso[k];
				
				/* sposta a sinistra di un campione il delay */
				delay_buffer[k-1] = delay_buffer[k]; 
			}
			
			/* aggiunge il nuovo campione in coda */
			delay_buffer[k-1] = 0; 
			
			/* scrive il campione risultante su file */
			WFW_Scrivi(wfo, &campione_output, 1);
		}
		free (delay_buffer);
	} /* M*M */

	/* x(n) mono, h(n) stereo */
	if (wfx->nChannels == 1 && canali_impulso == 2)
	{
		int campioni_da_leggere = wfx->nSamples;
		float  campione_input, 
			   campione_output[2];
			   
		
		float* delay_buffer = malloc(sizeof(float)*campioni_impulso);
		memset (delay_buffer, 0, sizeof(float)*campioni_impulso);
		
		while (campioni_da_leggere--)
		{
			WFD_LeggiCampione(wfx, &campione_input);
			campione_output[0] = delay_buffer[0] * impulso[0];
			campione_output[1] = delay_buffer[0] * impulso[1];

			for (k = 1; k < campioni_impulso; ++k)
			{	
				campione_output[0] += delay_buffer[k] * impulso[k*2];
				campione_output[1] += delay_buffer[k] * impulso[k*2+1];
				delay_buffer[k-1] = delay_buffer[k]; 
			}
			delay_buffer[k-1] = campione_input; 
			WFW_Scrivi(wfo, campione_output, 1);
		}
		
		for (j = 0; j < campioni_impulso; ++j)
		{
			campione_output[0] = delay_buffer[0] * impulso[0];
			campione_output[1] = delay_buffer[0] * impulso[1];
			for (k = 1; k < campioni_impulso; ++k)
			{	
				campione_output[0] += delay_buffer[k] * impulso[k*2];
				campione_output[1] += delay_buffer[k] * impulso[k*2+1];
				delay_buffer[k-1] = delay_buffer[k]; 
			}

			delay_buffer[k-1] = 0; 
			WFW_Scrivi(wfo, campione_output, 1);
		}
		free(delay_buffer);
	} /* M*S */
	
	/* x(n) stereo, h(n) stereo */
	if (wfx->nChannels == 2 && canali_impulso == 2)
	{
		int campioni_da_leggere = wfx->nSamples;
		float  campione_input[2], 
			   campione_output[2];
		
		
		float* delay_buffer = malloc(sizeof(float)*campioni_impulso*2);
		memset (delay_buffer, 0, sizeof(float)*campioni_impulso*2);
		
		while (campioni_da_leggere--)
		{
			WFD_LeggiCampione(wfx, campione_input);
			campione_output[0] = delay_buffer[0] * impulso[0];
			campione_output[1] = delay_buffer[1] * impulso[1];
			
			for (k = 1; k < campioni_impulso; ++k)
			{	
				campione_output[0] += delay_buffer[k*2] * impulso[k*2];
				campione_output[1] += delay_buffer[k*2+1] * impulso[k*2+1];
				delay_buffer[(k-1)*2] = delay_buffer[k*2]; 
				delay_buffer[(k-1)*2+1] = delay_buffer[k*2+1]; 
			}

			delay_buffer[(k-1)*2] = campione_input[0];
			delay_buffer[(k-1)*2+1] = campione_input[1]; 
			
			WFW_Scrivi(wfo, campione_output, 1);
		}
		
		for (j = 0; j < campioni_impulso; ++j)
		{
			campione_output[0] = delay_buffer[0] * impulso[0];
			campione_output[1] = delay_buffer[1] * impulso[1];

			for (k = 1; k < campioni_impulso; ++k)
			{	
				campione_output[0] += delay_buffer[k*2] * impulso[k*2];
				campione_output[1] += delay_buffer[k*2+1] * impulso[k*2+1];
				delay_buffer[(k-1)*2] = delay_buffer[k*2]; 
				delay_buffer[(k-1)*2+1] = delay_buffer[k*2+1]; 
			}

			delay_buffer[(k-1)*2] = 0;
			delay_buffer[(k-1)*2+1] = 0;
			
			WFW_Scrivi(wfo, campione_output, 1);
		}

		free (delay_buffer);
	} /* S*S */

	/* x(n) stereo, h(n) mono*/
	if (wfx->nChannels == 2 && canali_impulso == 1)
	{
		int campioni_da_leggere = wfx->nSamples;
		float  campione_input[2], 
			   campione_output[2];
		
		
		float* delay_buffer = malloc(sizeof(float)*campioni_impulso*2);
		memset (delay_buffer, 0, sizeof(float)*campioni_impulso*2);
		
		while (campioni_da_leggere--)
		{
			WFD_LeggiCampione(wfx, campione_input);
			campione_output[0] = delay_buffer[0] * impulso[0];
			campione_output[1] = delay_buffer[1] * impulso[0];
			
			for (k = 1; k < campioni_impulso; ++k)
			{	
				campione_output[0] += delay_buffer[k*2] * impulso[k];
				campione_output[1] += delay_buffer[k*2+1] * impulso[k];
				delay_buffer[(k-1)*2] = delay_buffer[k*2]; 
				delay_buffer[(k-1)*2+1] = delay_buffer[k*2+1]; 
			}
			
			delay_buffer[(k-1)*2] = campione_input[0];
			delay_buffer[(k-1)*2+1] = campione_input[1]; 
			
			WFW_Scrivi(wfo, campione_output, 1);
		}
		
		for (j = 0; j < campioni_impulso; ++j)
		{
			campione_output[0] = delay_buffer[0] * impulso[0];
			campione_output[1] = delay_buffer[1] * impulso[0];
			
			for (k = 1; k < campioni_impulso; ++k)
			{	
				campione_output[0] += delay_buffer[k*2] * impulso[k];
				campione_output[1] += delay_buffer[k*2+1] * impulso[k];
				delay_buffer[(k-1)*2] = delay_buffer[k*2]; 
				delay_buffer[(k-1)*2+1] = delay_buffer[k*2+1]; 
			}
			
			delay_buffer[(k-1)*2] = 0;
			delay_buffer[(k-1)*2+1] = 0;
			
			WFW_Scrivi(wfo, campione_output, 1);
		}
		
		free (delay_buffer);
	} /* S*M */	
}

/* Carica l'impulso in memoria RAM */
float* carica_impulso (WaveFileDirect* wfh)
{
	float *h = 0;	
	if (wfh == NULL) goto cleanup;

	/* acquisisce memoria per contenere tutto l'impulso */
	h = malloc (wfh->nSamples * sizeof(float) * wfh->nChannels);	
	if (h == NULL) goto cleanup;	

	/* Legge l'impulso dal file wave */
	if (WFD_Leggi(wfh, h, wfh->nSamples) == 0) goto cleanup;

	/* ritorna il puntatore all'impulso */
	return h;

cleanup: /* in caso di errore... */
	
	if (h != NULL) free(h);

	return NULL;
}

void usage ()
{
	printf ("-------------------- Convolver -------------------\n");
	printf ("   (C) 2001 Aldo Trianti - per Cubase Magazine\n");
	printf ("--------------------------------------------------\n\n");
	printf ("Un semplice programma che effettua la convoluzione\n"); 
	printf ("tra due file wave.\n");
	printf ("\n");
	printf ("Utilizzo del programma a linea di comando:\n\n");
	printf ("conv input h output\n\n");
	printf (" input = nome del file wave da processare\n");
	printf ("     h = nome del file wave contenente la risposta di impulso\n");
	printf ("output = nome del file wave che verra' prodotto (input**h)\n");
	exit(EXIT_FAILURE);
	
}


void giveup (char* why)
{
	printf ("Errore: %s\n", why);
	exit(EXIT_FAILURE);
}

