#include <stdio.h>
#include <stdlib.h>

#include "wavereader.h"


/*---------------------------------------------------------------------------*/
int WFW_Chiudi(WaveFileWriter* wfw)
{
	int RIFFchunkSize;
	int dataChunkSize;
	int offset = ftell(wfw->file);
		

	/* il chunk RIFF  grande quanto tutto il file wave meno lo header
	   del chunk stesso */
	RIFFchunkSize = offset - 12;

	/* scrive la dimensione del chunk */
	fseek (wfw->file, 4, SEEK_SET);
	fwrite(&offset, sizeof(offset), 1, wfw->file);

	dataChunkSize = wfw->nSamples * 2 * wfw->nChannels;

	/* scrive la dimensione dei dati PCM nel chunk 'data' */
	fseek(wfw->file, offset - dataChunkSize - 4, SEEK_SET);
	fwrite(&dataChunkSize, 4, 1, wfw->file);		

	return 0;
};


/*---------------------------------------------------------------------------*/
int WFW_Scrivi(WaveFileWriter *wfw, float* buffer, int sampleframes)
{
	short temp[1024];
	int k;
	float scale = 32768;	

	int samples = sampleframes * wfw->nChannels;

	while (samples)
	{
		int daScrivere = samples;
		if (daScrivere > 1024) daScrivere = 1024;

		for (k = 0; k < daScrivere; ++k)		
		{		
			int sample = buffer[k] * 32768;
			if (sample > 32767) sample = 32767;
			if (sample < -32768) sample = -32768;
			temp[k] = sample;
		}
		
		if (fwrite(temp, sizeof(short), daScrivere, wfw->file) 
			!= daScrivere) break;

		samples -= daScrivere;
		buffer += daScrivere;		
	}

	if (samples == 0)
	{	
		wfw->nSamples += sampleframes;
		return 1;
	}
	else
		return 0;
}


/*---------------------------------------------------------------------------*/
WaveFileWriter* WFW_CreaWave(char *nome_file, int sr, int canali)
{
	WaveFileWriter *wfw = malloc(sizeof(*wfw));
	if (wfw)
	{
		RiffChunk riff;
		unsigned int waveType = TYPE_WAVE;
		RiffChunk fmt;
		RiffChunk data;
		WaveFormat wf;
		int check;
		
		FILE* f = fopen (nome_file, "wb");
		if (f == NULL) goto cleanup;

		wfw->file = f;
		wfw->nChannels = canali;
		wfw->sampleRate = sr;
		wfw->nSamples = 0;

		/* scrive chunk RIFF */
		riff.id = RIFF_CHUNK;
		riff.cbSize = 0; /* Da determinarsi in seguito */
		check = fwrite (&riff, 1, sizeof riff, f);
		if (check != sizeof riff) goto cleanup;

		check = fwrite (&waveType, 1, sizeof waveType, f);
		if (check != sizeof waveType) goto cleanup;

		/* scrive intestazione chunk FMT (formato wave) */
		fmt.id = FMT__CHUNK;
		fmt.cbSize = sizeof(wf);
		check = fwrite (&fmt, 1, sizeof fmt, f);
		if (check != sizeof fmt) goto cleanup;

		/* scrive il formato vero e proprio */
		wf.nAvgBytesPerSec = 2 * sr * canali;
		wf.nBlockAlign = 2 * canali;
		wf.nChannels = canali;
		wf.nSamplesPerSec = sr;
		wf.wBitsPerSample = 16;
		wf.wFormatTag = WAVE_FORMAT_PCM;
		check = fwrite(&wf, 1, sizeof wf, f);
		if (check != sizeof wf) goto cleanup;

		/* scrive intestazione dati PCM */
		data.id = DATA_CHUNK;
		data.cbSize = 0; /* Da determinarsi in seguito */
		check = fwrite(&data, 1, sizeof data, f);
		if (check != sizeof data) goto cleanup;
	}
	return wfw;

	cleanup:
	if (wfw->file)
		fclose(wfw->file);

	if (wfw != NULL)
		free(wfw);

	return NULL;
}

/* #Definire TESTME per generare un programma di prova */
#ifdef TESTME
#include <math.h>

int main()
{
	float* data;
	int k;
	double pi, omega;
	WaveFileWriter* wfw;
	
	pi = atan(1)*4;
	omega = 1200. / 44100. * 2 * pi; /* tono a 1.2 kHz */
	
	wfw = WFW_CreaWave("c:\\test.wav", 44100, 2);
	
	data = malloc(65536*sizeof(float));
	
	for (k = 0; k < 32768; ++k) 
	{
		data[k*2] = sin(k*omega);
		data[k*2+1] = cos(k*omega);
	}
	
	WFW_Scrivi(wfw, data, 32768);
	WFW_Chiudi(wfw);
	
	return 0;
}

#endif
